package path;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import box.Sector;

public class Area {
	private ArrayList<ArrayList<Node>> nodeMap;
	private List<Node> open, closed;
	private int width, height;
	
	//DEBUG
	//There are four sectors, sectors A, B, C, and D.
	private ArrayList<Sector> sectors;
	//DEBUG
	//out of memory errors prevention
	private ArrayList<Node> tempPath;
	
	/*
	     |
	   A | B       
	 ____+___  
	     |      
	   C | D    
	         
	   
	 * */
	
	public Area(int w, int h) {
		nodeMap = new ArrayList<ArrayList<Node>>();
		Random random = new Random();
		for (int x = 0; x < w; x++) {
			ArrayList<Node> temp = new ArrayList<Node>();
			for (int y = 0; y < h; y++) {
				Node node = new Node(x, y);
				if (Math.floor(random.nextDouble() + 0.15) >= 0.8)
					node.setNodeAsObstacle();
				temp.add(node);
			}
			nodeMap.add(temp);
		}
		
		open = new ArrayList<Node>();
		closed = new ArrayList<Node>();
		width = w;
		height = h;
		
		initializeSectors();
	}
	
	public ArrayList<ArrayList<Node>> getNodeMap() {
		return nodeMap;
	}
	
	public ArrayList<Node> makePath(Node start, Node goal) {
		System.out.println("Making path from sectors...");
		if (tempPath == null)
			tempPath = new ArrayList<Node>();
		Sector startingSector = findSector(start);
		Sector endingSector = findSector(goal);
		
		if (startingSector == null || endingSector == null) {
			System.out.println("Something is wrong...");
			return null;
		}
		
		System.out.println("Calculating different segments of paths from each sectors...");
		System.out.println("The first path...");
		tempPath = createPath(start, startingSector.centerNode);
		System.out.println("The second path...");
		tempPath.addAll(createPath(startingSector.centerNode, endingSector.centerNode));
		System.out.println("The last path...");
		tempPath.addAll(createPath(endingSector.centerNode, goal));
		System.out.println("Paths complete.");
		return tempPath;
	}
	
	private Sector findSector(Node node) {
		for (Sector sector : sectors) {
			if (sector.boundingBox.hasNode(node))
				return sector;
		}
		return null;
	}
	
	private void initializeSectors() {
		System.out.println("Initializing sectors.");
		int halfWidth = width / 2;
		int halfHeight = height / 2;
		sectors = new ArrayList<Sector>();
		
		for (int i = 0; i < 2; i++) {
			for (int j = 0; j < 2; j++) {
				Sector sector = new Sector(j * halfWidth, i * halfHeight, halfWidth, halfHeight);
				sectors.add(sector);
			}
		}
		
		//Order of operations
		Sector sector = sectors.get(0);
		sector.setPortal(sectors.get(1));
		sector.setPortal(sectors.get(2));
		sector = sectors.get(1);
		sector.setPortal(sectors.get(0));
		sector.setPortal(sectors.get(3));
		sector = sectors.get(2);
		sector.setPortal(sectors.get(0));
		sector.setPortal(sectors.get(3));
		sector = sectors.get(3);
		sector.setPortal(sectors.get(1));
		sector.setPortal(sectors.get(2));
		
		System.out.println("Sectors complete.");
	}
	
	public ArrayList<Node> createPath(Node start, Node goal) {
		start.distanceFromStart = 0;
		start.parent = null;
		start.hueristicDistanceToGoal = 0;
		goal.parent = null;
		goal.distanceFromStart = getEstimatedDistanceToGoal(start, goal);
		goal.hueristicDistanceToGoal = getDistanceBetween(start, goal);
		open.clear();
		closed.clear();
		open.add(start);
		Node current = null;
		while (!open.isEmpty()) {
			Collections.sort(open);
			current = open.remove(0);
			if (current.equals(goal)) {
				goal.parent = current;
				goal.distanceFromStart = current.distanceFromStart + getDistanceBetween(current, goal);
				goal.hueristicDistanceToGoal = 0;
				break;
			}
			open.remove(current);
			closed.add(current);
			for (Node neighbor : getNeighbors(current)) {
				boolean neighborIsBetter;
				if (closed.contains(neighbor))
					continue;
				double neighborDistanceFromStart = current.distanceFromStart + getDistanceBetween(current, neighbor);
				double neighborDistanceToGoal = getEstimatedDistanceToGoal(neighbor, goal);
				double currentDistanceToGoal = getEstimatedDistanceToGoal(current, goal);
				if (neighborDistanceToGoal >= currentDistanceToGoal)
					continue;
				if (!neighbor.isObstacle()) {
					if (!open.contains(neighbor)) {
						open.add(neighbor);
						Collections.sort(open);
						neighborIsBetter = true;
					}
					else if (neighborDistanceFromStart < current.distanceFromStart)
						neighborIsBetter = true;
					else
						neighborIsBetter = false;
					if (neighborIsBetter) {
						neighbor.parent = current;
						neighbor.distanceFromStart = neighborDistanceFromStart;
						neighbor.hueristicDistanceToGoal = getEstimatedDistanceToGoal(neighbor, goal);
					}
				}
				
			}
		}
		if (current != null) {
			System.out.println("Successfully created a path, now reconstructing...");
			return recreatePath(goal);
		}
		System.out.println("ABORT - System failed......");
		return null;
	}
	
	private ArrayList<Node> getNeighbors(Node node) {
		ArrayList<Node> neighborList = new ArrayList<Node>();
		int x = node.x;
		int y = node.y;
		
		int x0 = x - 1;
		int x2 = x + 1;
		int y0 = y - 1;
		int y2 = y + 1;
		
		if (x < 0 || y < 0 || x > width - 1 || y > height - 1)
			return null;
		
		//----------------------------------------------
		if (x0 > 0 && y0 > 0)
			neighborList.add(nodeMap.get(x0).get(y0));
		if (x0 > 0)
			neighborList.add(nodeMap.get(x0).get(y));
		if (x0 > 0 && y2 < height - 1)
			neighborList.add(nodeMap.get(x0).get(y2));
		//----------------------------------------------
		if (y0 > 0)
			neighborList.add(nodeMap.get(x).get(y0));
		if (y2 < height - 1)
			neighborList.add(nodeMap.get(x).get(y2));
		//----------------------------------------------
		if (x2 < width - 1 && y0 > 0)
			neighborList.add(nodeMap.get(x2).get(y0));
		if (x2 < width - 1)
			neighborList.add(nodeMap.get(x2).get(y));
		if (x2 < width - 1 && y2 < height - 1)
			neighborList.add(nodeMap.get(x2).get(y2));
		//----------------------------------------------
		
		return neighborList;
	}
	
	private ArrayList<Node> getCardinalNeighbors(Node node) {
		ArrayList<Node> neighborList = new ArrayList<Node>();
		int x = node.x;
		int y = node.y;
		
		int x0 = x - 1;
		int x2 = x + 1;
		int y0 = y - 1;
		int y2 = y + 1;
		
		if (x < 0 || y < 0 || x > width - 1 || y > height - 1)
			return null;
		
		//----------------------------------------------
		if (x0 > 0)
			neighborList.add(nodeMap.get(x0).get(y));
		//----------------------------------------------
		if (y0 > 0)
			neighborList.add(nodeMap.get(x).get(y0));
		if (y2 < height - 1)
			neighborList.add(nodeMap.get(x).get(y2));
		//----------------------------------------------
		if (x2 < width - 1)
			neighborList.add(nodeMap.get(x2).get(y));
		//----------------------------------------------
		
		return neighborList;
	}
	
	private double getDistanceBetween(Node n1, Node n2) {
		if (n1.x == n2.x || n1.y == n2.y)
			return 1.0;
		else
			//Diagonal distance from node 1 to node 2.
			return 1.7;
	}
	
	private double getEstimatedDistanceToGoal(Node start, Node finish) {
		double h_diagonal = Math.min(Math.abs(start.x - finish.x), Math.abs(start.y - finish.y));
		double h_straight = Math.abs(start.x - finish.x) + Math.abs(start.y - finish.y);
		double h_result = Math.sqrt(2) * h_diagonal + (h_straight - 2 * h_diagonal);
		
		//breaking ties
		double p = 1 / 10000;
		h_result *= (1.0 + p);
		return h_result;
	}
	
	private ArrayList<Node> recreatePath(Node node) {
		int counter = 0;
		ArrayList<Node> results = new ArrayList<Node>();
		while (node.parent != null) {
			if (node.equals(node.parent)) {
				counter++;
				if (counter > 15)
					break;
			}
			results.add(node);
			node = node.parent;
		}
		//DEBUG
		System.out.println("Size of path: " + results.size());
		return results;
	}
}
